@@ -135,6 +135,11 @@ static apr_status_t serf_dechunk_read(serf_bucket_t *bucket,
                 ctx->body_left = 2;     /* CRLF */
             }
 
+            /* We need more data but there is no more available. */
+            if (ctx->body_left && APR_STATUS_IS_EOF(status)) {
+                return SERF_ERROR_TRUNCATED_HTTP_RESPONSE;
+            }
+
             /* Return the data we just read. */
             return status;
 
@@ -148,10 +153,18 @@ static apr_status_t serf_dechunk_read(serf_bucket_t *bucket,
              * if we're done reading the chunk terminator.
              */
             ctx->body_left -= *len;
+
+            /* We need more data but there is no more available. */
+            if (ctx->body_left && APR_STATUS_IS_EOF(status))
+                return SERF_ERROR_TRUNCATED_HTTP_RESPONSE;
+
             if (!ctx->body_left) {
                 ctx->state = STATE_SIZE;
             }
 
+            /* Don't return the CR of CRLF to the caller! */
+            *len = 0;
+
             if (status)
                 return status;
 
@@ -159,6 +172,7 @@ static apr_status_t serf_dechunk_read(serf_bucket_t *bucket,
 
         case STATE_DONE:
             /* Just keep returning EOF */
+            *len = 0;
             return APR_EOF;
 
         default:
